/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * Classes to deal with multi-arity operators (functions with the attributes
 * :left_assoc, :right_assoc, :chainable)
 */

#ifndef MULTI_ARITY_OPS
#define MULTI_ARITY_OPS 1

#include "smtlib-resolver.H"

namespace Resolver {

class MultiArityFunctionFamily : public FunctionFamily {
private:
    // We just use a vector to store the created instances
    // TODO: replace this with a more efficient map
    std::vector<MonoFunctionPtr> instances;

protected:
    const FunctionFamilyPtr basisFunction;

    virtual MonoFunctionPtr
        createInstanceFor(const std::vector<SortPtr>& argumentSorts) = 0;

public:
    MultiArityFunctionFamily(const FunctionFamilyPtr& basisFunction);
    MultiArityFunctionFamily(const MonoFunctionPtr& basisFunction);

    MonoFunctionPtr
        lookup(const std::vector<std::string>& indexes,
               const std::vector<SortPtr>& argumentSorts);
    MonoFunctionPtr
        lookup(const std::vector<std::string>& indexes,
               const SortPtr& resSort,
               const std::vector<SortPtr>& argumentSorts);
};

typedef CountingPtr<MultiArityFunctionFamily> MultiArityFunctionFamilyPtr;

class LeftAssocFunctionFamily : public MultiArityFunctionFamily {
protected:
    MonoFunctionPtr createInstanceFor(const std::vector<SortPtr>& argumentSorts);

public:
    LeftAssocFunctionFamily(const FunctionFamilyPtr& basisFunction);
    LeftAssocFunctionFamily(const MonoFunctionPtr& basisFunction);
};

class RightAssocFunctionFamily : public MultiArityFunctionFamily {
protected:
    MonoFunctionPtr createInstanceFor(const std::vector<SortPtr>& argumentSorts);

public:
    RightAssocFunctionFamily(const FunctionFamilyPtr& basisFunction);
    RightAssocFunctionFamily(const MonoFunctionPtr& basisFunction);
};

class ChainableFunctionFamily : public MultiArityFunctionFamily {
private:
    const MonoFunctionPtr fun_true;
    const MonoFunctionPtr fun_and;

protected:
    MonoFunctionPtr createInstanceFor(const std::vector<SortPtr>& argumentSorts);

public:
    ChainableFunctionFamily(const FunctionFamilyPtr& basisFunction,
                            const MonoFunctionPtr& fun_true,
                            const MonoFunctionPtr& fun_and);
    ChainableFunctionFamily(const MonoFunctionPtr& basisFunction,
                            const MonoFunctionPtr& fun_true,
                            const MonoFunctionPtr& fun_and);
};

////////////////////////////////////////////////////////////////////////////////

class ExpandableMonoFunction : public MonoFunction {
protected:
    const FunctionFamilyPtr basisFunction;

public:
    template <class I> ExpandableMonoFunction(const IndexedString& name,
                                              SortPtr res, I argBegin, I argEnd,
                                              const FunctionFamilyPtr& basisFunction)
        : MonoFunction(name, res, argBegin, argEnd), basisFunction(basisFunction) {}

    virtual ExpressionPtr expand(const std::vector<ExpressionPtr>& arguments) = 0;
};

typedef CountingPtr<ExpandableMonoFunction> ExpandableMonoFunctionPtr;

}

#endif
