/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "relation-axioms.H"

namespace SMTBackend {

using namespace std;

RelationAxiomGenerator::RelationAxiomGenerator(std::ostream& os)
    : os(os) {}

void RelationAxiomGenerator::tuple(const SortPtr& sort) {
    os << "tuple_" << sort->op->arity << "_";
    idLineariser.linearise(os, sort);
}

void RelationAxiomGenerator::project(const SortPtr& sort, unsigned int index) {
    os << "project_" << sort->op->arity << "_" << index << "_";
    idLineariser.linearise(os, sort);
}

void RelationAxiomGenerator::product(const SortPtr& sort) {
    os << "product_" << sort->arguments.front()->op->arity << "_";
    idLineariser.linearise(os, sort);
}

void RelationAxiomGenerator::startComment(const SortPtr& sort,
                                          const std::string& desc) {
    os << ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;"
       << endl;
    os << ";; Instance ";
    idLineariser.linearise(os, sort);
    os << " of " << desc << endl << endl;
}

void RelationAxiomGenerator::stopComment(const SortPtr& sort,
                                         const std::string& desc) {
    os << endl;
    os << ";; End: Instance ";
    idLineariser.linearise(os, sort);
    os << " of " << desc << endl;
    os << ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;"
       << endl << endl;
}

void RelationAxiomGenerator::instantiateTupleAxioms(const SortPtr& sort) {
    startComment(sort, "theory of finite relations with cardinality");

    // the tuple sort
    os << ":extrasorts (";
    idLineariser.linearise(os, sort);
    os << ")" << endl;

    // the tuple function symbols (constructor and projections)
    os << ":extrafuns (";

    os << "(";
    tuple(sort);
    for (vector<SortPtr>::const_iterator it = sort->arguments.begin();
         it != sort->arguments.end();
         ++it) {
        os << " ";
        idLineariser.linearise(os, *it);
    }
    os << " ";
    idLineariser.linearise(os, sort);
    os << ")" << endl;

    for (unsigned int i = 1; i <= sort->op->arity; ++i) {
        os << "(";
        project(sort, i);
        os << " ";
        idLineariser.linearise(os, sort);
        os << " ";
        idLineariser.linearise(os, sort->arguments[i-1]);
        os << ")" << endl;
    }

    os << ")" << endl;

    // the axioms: projections are the left-inverse of the tuple constructor
    for (unsigned int j = 1; j <= sort->op->arity; ++j) {
        os << ":assumption" << endl;
        os << "(forall";

        for (unsigned int i = 1; i <= sort->op->arity; ++i) {
            os << " (?x" << i << " ";
            idLineariser.linearise(os, sort->arguments[i-1]);
            os << ")";
        }

        os << " (= (";
        project(sort, j);
        os << " (";
        tuple(sort);
        for (unsigned int i = 1; i <= sort->op->arity; ++i)
            os << " ?x" << i;
        os << ")) ?x" << j << "))";

        os << endl;
    }

    // the axioms: and the tuple constructor is the left-inverse of the
    // projections
    os << ":assumption" << endl;
    os << "(forall (?x ";
    idLineariser.linearise(os, sort);
    os << ") (= (";
    tuple(sort);

    for (unsigned int i = 1; i <= sort->op->arity; ++i) {
        os << " (";
        project(sort, i);
        os << " ?x)";
    }
    os << ") ?x)";
    os << ")" << endl;

    stopComment(sort, "theory of finite relations with cardinality");
}

void RelationAxiomGenerator::instantiateSetTupleAxioms(const SortPtr& sort) {
    startComment(sort, "cartesian products");

    const SortPtr tupleSort = sort->arguments.front();

    // the product function symbol
    os << ":extrafuns ((";
    product(sort);
    for (unsigned int i = 0; i < tupleSort->op->arity; ++i) {
        os << " Set_";
        idLineariser.linearise(os, tupleSort->arguments[i]);
        os << "";
    }
    os << " ";
    idLineariser.linearise(os, sort);
    os << "))" << endl;

    // the axiom for cartesian products
    os << ":assumption" << endl;
    os << "(forall";
    for (unsigned int i = 1; i <= tupleSort->op->arity; ++i) {
        os << " (?x" << i << " Set_";
        idLineariser.linearise(os, tupleSort->arguments[i-1]);
        os << ")";
    }
    os << " (?y ";
    idLineariser.linearise(os, tupleSort);
    os << ")";

    os << " (iff (in_";
    idLineariser.linearise(os, sort);
    os << " ?y (";
    product(sort);
    for (unsigned int i = 1; i <= tupleSort->op->arity; ++i)
        os << " ?x" << i;
    os << ")) ";
    os << " (and";
    for (unsigned int i = 1; i <= tupleSort->op->arity; ++i) {
        os << " (in_Set_";
        idLineariser.linearise(os, tupleSort->arguments[i-1]);
        os << " (";
        project(tupleSort, i);
        os << " ?y) ?x" << i << ")";
    }
    os << ")))";

    stopComment(sort, "cartesian products");
}

}
