/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * Datastructures to represent SMT-LIB expressions and benchmarks
 *
 * TODO: annotations
 */

#ifndef SMTLIB_ABSYN
#define SMTLIB_ABSYN 1

#include <string>
#include <vector>
#include <iostream>

#include "refcounting.H"

namespace SMTLibAbsyn {

// Forward definitions
class Sort;
class SortCtor;
class Expression;
class MonoFunction;
class FunctionApp;
class Variable;
class Literal;
class LetBinder;
class QuantifiedExpression;

typedef CountingPtr<Sort> SortPtr;
typedef CountingPtr<SortCtor> SortCtorPtr;
typedef CountingPtr<Expression> ExpressionPtr;
typedef CountingPtr<Variable> VariablePtr;
typedef CountingPtr<FunctionApp> FunctionAppPtr;
typedef CountingPtr<MonoFunction> MonoFunctionPtr;

/**
 * Names in SMT-LIB 2 consist of a string + a sequence of indices (numerals)
 */
struct IndexedString {
	IndexedString(const char* body);

	IndexedString(const std::string& body,
	              const std::vector<std::string>& indexes =
	            		  std::vector<std::string>());

	std::string body;
	std::vector<std::string> indexes;

	std::string toString() const;

	bool operator<(const IndexedString& that) const;
};

std::ostream& operator<< (std::ostream& os, const IndexedString& str);

////////////////////////////////////////////////////////////////////////////////
// Sort ASTs

/**
 * Sort constructors are modelled as unique objects
 */
class SortCtor : public RefCounted {

public:
	SortCtor(const IndexedString& name, unsigned int arity = 0)
		: name(name), arity(arity) {}

	SortPtr m ();
	SortPtr m (const SortPtr& arg0);
	SortPtr m (const SortPtr& arg0, const SortPtr& arg1);

	const IndexedString name;
	const unsigned int arity;

	/**
	 * We store the boolean sort globally, because it has a distinguished role
	 * in the expression ASTs (the type of quantified expressions is always
	 * Bool). Really, Bool belongs to the empty theory.
	 */
	static const SortCtorPtr Bool;
};

std::ostream& operator<< (std::ostream& os, const SortCtor& c);



/**
 * Sorts are terms over the sort constructors
 */
class Sort : public RefCounted {

public:
	Sort(const SortCtorPtr& op);
	template <class I> Sort(const SortCtorPtr& op, I begin, I end)
		: op(op), arguments(begin, end) {
		assert(op->arity == arguments.size());
	}
	Sort(const SortCtorPtr& op, const SortPtr& arg0);
	Sort(const SortCtorPtr& op, const SortPtr& arg0, const SortPtr& arg1);

	const SortCtorPtr op;
	const std::vector<SortPtr> arguments;

	bool operator==(const Sort& that) const;

	SortPtr update(const std::vector<SortPtr>& newArguments);

	static const SortPtr Bool;
};

std::ostream& operator<< (std::ostream& os, const Sort& c);

////////////////////////////////////////////////////////////////////////////////
// Expression ASTs

/**
 * A monomorphic function. Polymorphic functions are modelled as families of
 * monomorphic functions (the classes FunctionFamily and PolyFunctionFamily)
 */
class MonoFunction : public RefCounted {

public:
	template <class I> MonoFunction(const IndexedString& name,
					                SortPtr res, I argBegin, I argEnd)
		: argumentSorts(argBegin, argEnd), name(name), resultSort(res) {}

	MonoFunction(const std::string& name, SortPtr res);

    MonoFunction(const IndexedString& name,
                 SortPtr res, const std::vector<SortPtr>& argumentSorts);

	const std::vector<SortPtr> argumentSorts;
	const IndexedString name;
	const SortPtr resultSort;

	ExpressionPtr m ();
	ExpressionPtr m (const ExpressionPtr& arg0);
	ExpressionPtr m (const ExpressionPtr& arg0,
	                 const ExpressionPtr& arg1);
    ExpressionPtr m (const ExpressionPtr& arg0,
                     const ExpressionPtr& arg1,
                     const ExpressionPtr& arg2);

	template <class I>
	ExpressionPtr m (I argBegin, I argEnd) {
	    return new FunctionApp(this, argBegin, argEnd);
	}
};

std::ostream& operator<< (std::ostream& os, const MonoFunction& f);

/**
 * Visitor for the expression AST datastructure. Further visitors and subclasses
 * are defined in smtlib-absyn-visitors.H
 */
class ExpressionVisitor {
public:
	virtual void visitVariable            (const CountingPtr<Variable>& node) = 0;
	virtual void visitLiteral             (const CountingPtr<Literal>& node) = 0;
	virtual void visitFunctionApp         (const CountingPtr<FunctionApp>& node) = 0;
	virtual void visitLetBinder           (const CountingPtr<LetBinder>& node) = 0;
	virtual void visitQuantifiedExpression(const CountingPtr<QuantifiedExpression>& node) = 0;
};

/**
 * The two kinds of quantifiers available in SMT-LIB
 */
enum Quantifier { Forall, Exists };

/**
 * Base class of all expressions
 */
class Expression : public RefCounted {

public:

	virtual SortPtr sort() const;
	virtual unsigned int arity() const;
    virtual ExpressionPtr subExpr(unsigned int index) const;
	virtual void visit(ExpressionVisitor& visitor);

	/**
	 * Update the sub-expressions of this expression, avoiding the creating of
	 * new objects if possible
	 */
	virtual ExpressionPtr update(const std::vector<ExpressionPtr>& newSubExpressions);
};

std::ostream& operator<< (std::ostream& os, Expression& expr);

std::vector<SortPtr> extractSorts(const std::vector<ExpressionPtr>& expressions);

/**
 * Bound variables (using de Brujin notation)
 */
class Variable : public Expression {

private:
	const SortPtr _sort;

public:
	Variable(const SortPtr& s, unsigned int index);

	const unsigned int index;

	SortPtr sort() const;
	void visit(ExpressionVisitor& visitor);
};

CountingPtr<Variable> v(const SortPtr& s, unsigned int index);

/**
 * Various literals supported in SMT-LIB. The value of a literal is just stored
 * as a string
 */
class Literal : public Expression {

private:
	const SortPtr _sort;

public:
	Literal(const SortPtr& s, const std::string& value);

	const std::string value;

	SortPtr sort() const;
	void visit(ExpressionVisitor& visitor);
};

/**
 * Function applications
 */
class FunctionApp : public Expression {

public:
	template <class I> FunctionApp(const MonoFunctionPtr& function,
					               I argBegin, I argEnd)
		: function(function), arguments(argBegin, argEnd) {
		// check that the arguments are correctly typed
		assert(arguments.size() == function->argumentSorts.size());
		for (unsigned int i = 0; i < arguments.size(); ++i)
			assert(*arguments[i]->sort() == *function->argumentSorts[i]);
	}

	const MonoFunctionPtr function;
	const std::vector<ExpressionPtr> arguments;

	SortPtr sort() const;
    unsigned int arity() const;
    ExpressionPtr subExpr(unsigned int index) const;
	void visit(ExpressionVisitor& visitor);
	ExpressionPtr update(const std::vector<ExpressionPtr>& newSubExpressions);
};

/**
 * Abstract superclass for the different kinds of expressions that bind
 * variables. Because de Brujin-indices are used, only the names and types of
 * variables are stored here
 */
class Binder : public Expression {
private:
    static void checkVariableTypes(const std::vector<SortPtr>& variableSorts,
                                   const ExpressionPtr& body);

public:
	template <class I, class J>
	Binder(I namesBegin, I namesEnd, J sortsBegin, J sortsEnd,
 	       const ExpressionPtr& body)
		: variableNames(namesBegin, namesEnd), variableSorts(sortsBegin, sortsEnd),
		  body(body) {
		assert(variableNames.size() == variableSorts.size() && !variableNames.empty());
		checkVariableTypes(variableSorts, body);
	}

	/**
	 * The names and sorts of the bound variables, starting with the outermost
	 * one
	 */
	const std::vector<std::string> variableNames;
	const std::vector<SortPtr> variableSorts;
	const ExpressionPtr body;
};

/**
 * Parallel let-expressions
 */
class LetBinder : public Binder {

public:
	template <class I, class J, class K>
	LetBinder(I namesBegin, I namesEnd, J sortsBegin, J sortsEnd,
			  K termsBegin, K termsEnd, const ExpressionPtr& body)
		: Binder(namesBegin, namesEnd, sortsBegin, sortsEnd, body),
		  boundTerms(termsBegin, termsEnd) {
		assert(variableNames.size() == boundTerms.size());
		for (unsigned int i = 0; i < boundTerms.size(); ++i)
			assert(*variableSorts[i] == *boundTerms[i]->sort());
	}

	/**
	 * The bound terms
	 */
	const std::vector<ExpressionPtr> boundTerms;

	SortPtr sort() const;
    unsigned int arity() const;
    ExpressionPtr subExpr(unsigned int index) const;
	void visit(ExpressionVisitor& visitor);
    ExpressionPtr update(const std::vector<ExpressionPtr>& newSubExpressions);
};

/**
 * Quantified expressions
 */
class QuantifiedExpression : public Binder {

public:
	template <class I, class J>
	QuantifiedExpression(Quantifier quantifier,
			             I namesBegin, I namesEnd, J sortsBegin, J sortsEnd,
			             const ExpressionPtr& body)
		: Binder(namesBegin, namesEnd, sortsBegin, sortsEnd, body),
		  quantifier(quantifier) {
		assert(body->sort() == Sort::Bool);
	}

	QuantifiedExpression(Quantifier quantifier,
			             const std::string& name, const SortPtr& sort,
			             const ExpressionPtr& body);

	const Quantifier quantifier;

	SortPtr sort() const;
    unsigned int arity() const;
    ExpressionPtr subExpr(unsigned int index) const;
	void visit(ExpressionVisitor& visitor);
    ExpressionPtr update(const std::vector<ExpressionPtr>& newSubExpressions);
};

CountingPtr<QuantifiedExpression> forall(const std::string& name, const SortPtr& sort,
                                         const ExpressionPtr& body);
CountingPtr<QuantifiedExpression> exists(const std::string& name, const SortPtr& sort,
                                         const ExpressionPtr& body);

}

#endif
