/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <algorithm>
#include <sstream>

#include "smtlib-absyn.H"
#include "util.H"

using namespace std;

namespace SMTLibAbsyn {

IndexedString::IndexedString(const char* body)
	: body(body), indexes(vector<string>()) {}

IndexedString::IndexedString(const std::string& body,
                             const std::vector<std::string>& indexes)
	: body(body), indexes(indexes) {}

std::string IndexedString::toString() const {
	stringstream ss;
	ss << *this;
	return ss.str();
}

bool IndexedString::operator<(const IndexedString& that) const {
    return body < that.body ||
           body == that.body && indexes < that.indexes;
}

std::ostream& operator<< (std::ostream& os, const IndexedString& str) {
	if (str.indexes.empty()) {
		os << str.body;
	} else {
		os << "(ind " << str.body << " ";
		for_each(str.indexes.begin(), str.indexes.end(),
		         print<string>(os, " "));
		os << ")";
	}
	return os;
}

Sort::Sort(const SortCtorPtr& op) : op(op) {
	assert(op->arity == 0);
}

Sort::Sort(const SortCtorPtr& op, const SortPtr& arg0)
	: op(op), arguments(to_vector(arg0)) {
	assert(op->arity == 1);
}

Sort::Sort(const SortCtorPtr& op, const SortPtr& arg0, const SortPtr& arg1)
	: op(op), arguments(to_vector(arg0, arg1)) {
	assert(op->arity == 2);
}

bool Sort::operator==(const Sort& that) const {
	return op == that.op && equalVectors(arguments, that.arguments);
}

SortPtr Sort::update(const std::vector<SortPtr>& newArguments) {
    if (arguments == newArguments)
        return this;
    return new Sort(op, newArguments.begin(), newArguments.end());
}

ostream& operator<< (ostream& os, const Sort& c) {
	if (c.op->arity == 0) {
		os << c.op->name;
	} else {
		os << '(' << c.op->name << ' ';
		for_each(c.arguments.begin(), c.arguments.end(),
				 print_deref<SortPtr>(os, " "));
	    os << ')';
	}
	return os;
}

////////////////////////////////////////////////////////////////////////////////

SortPtr SortCtor::m() {
	return new Sort(this);
}
SortPtr SortCtor::m(const SortPtr& arg0) {
	return new Sort(this, arg0);
}
SortPtr SortCtor::m(const SortPtr& arg0, const SortPtr& arg1) {
	return new Sort(this, arg0, arg1);
}

ostream& operator<< (ostream& os, const SortCtor& c) {
	os << c.name << '/' << c.arity;
	return os;
}

const SortCtorPtr SortCtor::Bool (new SortCtor ("Bool"));

const SortPtr Sort::Bool (SortCtor::Bool->m());

////////////////////////////////////////////////////////////////////////////////

MonoFunction::MonoFunction(const std::string& name, SortPtr res)
	: argumentSorts(), name(name), resultSort(res) {}

MonoFunction::MonoFunction(const IndexedString& name,
                           SortPtr res, const std::vector<SortPtr>& argumentSorts)
    : argumentSorts(argumentSorts), name(name), resultSort(res) {}

ExpressionPtr MonoFunction::m () {
	vector<ExpressionPtr> empty;
	return new FunctionApp(this, empty.begin(), empty.end());
}
ExpressionPtr MonoFunction::m (const ExpressionPtr& arg0) {
	return new FunctionApp(this, &arg0, &arg0 + 1);
}
ExpressionPtr MonoFunction::m (const ExpressionPtr& arg0,
		                       const ExpressionPtr& arg1) {
	vector<ExpressionPtr> args;
	args.push_back(arg0);
	args.push_back(arg1);
	return new FunctionApp(this, args.begin(), args.end());
}

ExpressionPtr MonoFunction::m (const ExpressionPtr& arg0,
                               const ExpressionPtr& arg1,
                               const ExpressionPtr& arg2) {
    vector<ExpressionPtr> args;
    args.push_back(arg0);
    args.push_back(arg1);
    args.push_back(arg2);
    return new FunctionApp(this, args.begin(), args.end());
}

std::ostream& operator<< (std::ostream& os, const MonoFunction& f) {
	os << f.name << " : ";
	for_each(f.argumentSorts.begin(), f.argumentSorts.end(),
			 print_deref<SortPtr>(os, " x "));
	os << " -> " << *f.resultSort;
	return os;
}

////////////////////////////////////////////////////////////////////////////////

SortPtr Expression::sort() const {
    assert(false);
    return 0;
}

unsigned int Expression::arity() const {
    return 0;
}

ExpressionPtr Expression::subExpr(unsigned int index) const {
    assert(false);
    return 0;
}

void Expression::visit(ExpressionVisitor& visitor) {
    assert(false);
}

ExpressionPtr Expression::update(const std::vector<ExpressionPtr>& newSubExpressions) {
    assert(newSubExpressions.empty());
    return this;
}

Variable::Variable(const SortPtr& s, unsigned int index)
	: _sort(s), index(index) {}

SortPtr Variable::sort() const {
	return _sort;
}

void Variable::visit(ExpressionVisitor& visitor) {
	visitor.visitVariable(this);
}

CountingPtr<Variable> v(const SortPtr& s, unsigned int index) {
    return new Variable(s, index);
}

Literal::Literal(const SortPtr& s, const std::string& value)
	: _sort(s), value(value) {}

SortPtr Literal::sort() const {
	return _sort;
}

void Literal::visit(ExpressionVisitor& visitor) {
	visitor.visitLiteral(this);
}

SortPtr FunctionApp::sort() const {
	return function->resultSort;
}

void FunctionApp::visit(ExpressionVisitor& visitor) {
	visitor.visitFunctionApp(this);
}

unsigned int FunctionApp::arity() const {
    return function->argumentSorts.size();
}

ExpressionPtr FunctionApp::subExpr(unsigned int index) const {
    return arguments[index];
}

ExpressionPtr FunctionApp::update(const std::vector<ExpressionPtr>& newSubExpressions) {
    assert(newSubExpressions.size() == arity());
    if (arguments == newSubExpressions)
        return this;

    return new FunctionApp(function,
                           newSubExpressions.begin(), newSubExpressions.end());
}

SortPtr LetBinder::sort() const {
	return body->sort();
}

void LetBinder::visit(ExpressionVisitor& visitor) {
	visitor.visitLetBinder(this);
}

unsigned int LetBinder::arity() const {
    return boundTerms.size() + 1;
}

ExpressionPtr LetBinder::subExpr(unsigned int index) const {
    if (index == boundTerms.size())
        return body;
    else
        return boundTerms[index];
}

ExpressionPtr LetBinder::update(const std::vector<ExpressionPtr>& newSubExpressions) {
    assert(newSubExpressions.size() == arity());
    if (equal(boundTerms.begin(), boundTerms.end(), newSubExpressions.begin()) &&
        body == newSubExpressions.back())
        return this;

    return new LetBinder(variableNames.begin(), variableNames.end(),
                         variableSorts.begin(), variableSorts.end(),
                         newSubExpressions.begin(), newSubExpressions.end() - 1,
                         newSubExpressions.back());
}

QuantifiedExpression::QuantifiedExpression(Quantifier quantifier,
						                   const std::string& name, const SortPtr& sort,
						                   const ExpressionPtr& body)
	: Binder(&name, &name + 1, &sort, &sort + 1, body), quantifier(quantifier) {
	assert(body->sort() == Sort::Bool);
}

SortPtr QuantifiedExpression::sort() const {
	return Sort::Bool;
}

void QuantifiedExpression::visit(ExpressionVisitor& visitor) {
	visitor.visitQuantifiedExpression(this);
}

unsigned int QuantifiedExpression::arity() const {
    return 1;
}

ExpressionPtr QuantifiedExpression::subExpr(unsigned int index) const {
    assert(index == 0);
    return body;
}

ExpressionPtr QuantifiedExpression::update(const std::vector<ExpressionPtr>& newSubExpressions) {
    assert(newSubExpressions.size() == arity());
    if (body == newSubExpressions.front())
        return this;

    return new QuantifiedExpression(quantifier,
                                    variableNames.begin(), variableNames.end(),
                                    variableSorts.begin(), variableSorts.end(),
                                    newSubExpressions.front());
}

CountingPtr<QuantifiedExpression> forall(const std::string& name, const SortPtr& sort,
                                         const ExpressionPtr& body) {
    return new QuantifiedExpression(Forall, name, sort, body);
}

CountingPtr<QuantifiedExpression> exists(const std::string& name, const SortPtr& sort,
                                         const ExpressionPtr& body) {
    return new QuantifiedExpression(Exists, name, sort, body);
}

////////////////////////////////////////////////////////////////////////////////

std::vector<SortPtr> extractSorts(const std::vector<ExpressionPtr>& expressions) {
    vector<SortPtr> res;
    for (vector<ExpressionPtr>::const_iterator it = expressions.begin();
         it != expressions.end();
         ++it) {
        assert((*it)->sort());
        res.push_back((*it)->sort());
    }
    return res;
}

}
