/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <iostream>
#include <vector>
#include <string>

#include "smtlib-printer.H"
#include "smtlib-absyn.H"
#include "smtlib-absyn-visitors.H"


namespace SMTLibAbsyn {

using namespace std;
using namespace Visitors;

struct PrinterContext {
    PrinterContext(ostream& os) : os(os) {}

    ostream& os;
};

/**
 * Visitors for linearising SMT-LIB expressions. The string argument denotes
 * a separator that to be added to the output stream immediately before an
 * expression.
 */
class PrinterPreVisitor
	: public ArgExpressionVisitor<PrinterArgument, int, PreVisitResultPtr> {
private:
    PrinterContext& ctxt;

public:
    PrinterPreVisitor(PrinterContext& ctxt) : ctxt(ctxt) {}

    void visitVariable            (const CountingPtr<Variable>& node) {
        ctxt.os << arg0.prefix;
        ctxt.os << arg0.boundVariables->drop(node->index)->value;
        ctxt.os << "/" << node->index;
        res = keepArg();
    }

    void visitLiteral             (const CountingPtr<Literal>& node) {
        ctxt.os << arg0.prefix;
        ctxt.os << node->value;
        res = keepArg();
    }

    void visitFunctionApp         (const CountingPtr<FunctionApp>& node) {
        ctxt.os << arg0.prefix;
        if (node->arity() > 0)
            ctxt.os << "(";
        ctxt.os << node->function->name;
        // function arguments are separated by spaces
        res = uniSubArgs(PrinterArgument(arg0.boundVariables, " "));
    }

    void visitLetBinder           (const CountingPtr<LetBinder>& node) {
        ctxt.os << arg0.prefix << "(let (";

        vector<PrinterArgument> arguments;
        StringListPtr bodyBoundVars = arg0.boundVariables;

        for (unsigned int i = 0; i < node->variableNames.size(); ++i) {
        	const string& varName = node->variableNames[i];
            string sep = "";
            if (i > 0)
                sep = ") ";
            arguments.push_back(PrinterArgument(arg0.boundVariables,
                                                sep + "(" + varName + " "));
            bodyBoundVars = new SinglyLinkedList<string> (varName, bodyBoundVars);
        }
        arguments.push_back(PrinterArgument(bodyBoundVars, ") "));

        res = subArgs(arguments);
    }

    void visitQuantifiedExpression(const CountingPtr<QuantifiedExpression>& node) {
        ctxt.os << arg0.prefix;
        switch (node->quantifier) {
        case Forall:
            ctxt.os << "(forall (";
            break;
        case Exists:
            ctxt.os << "(exists (";
            break;
        }

        StringListPtr bodyBoundVars = arg0.boundVariables;

        for (unsigned int i = 0; i < node->variableNames.size(); ++i) {
        	const string& varName = node->variableNames[i];

        	if (i > 0)
                ctxt.os << " ";
            ctxt.os << "(" << varName << " " << *node->variableSorts[i] << ")";
            bodyBoundVars = new SinglyLinkedList<string> (varName, bodyBoundVars);
        }

        ctxt.os << ") ";

        res = uniSubArgs(PrinterArgument(bodyBoundVars, ""));
    }
};

class PrinterPostVisitor
	: public ArgExpressionVisitor<PrinterArgument, const std::vector<int>*, int> {
private:
    PrinterContext& ctxt;

public:
    PrinterPostVisitor(PrinterContext& ctxt) : ctxt(ctxt) {}

    void visitVariable            (const CountingPtr<Variable>& node) {}

    void visitLiteral             (const CountingPtr<Literal>& node) {}

    void visitFunctionApp         (const CountingPtr<FunctionApp>& node) {
        if (node->arity() > 0)
            ctxt.os << ")";
    }

    void visitLetBinder           (const CountingPtr<LetBinder>& node) {
        ctxt.os << ")";
    }

    void visitQuantifiedExpression(const CountingPtr<QuantifiedExpression>& node) {
        ctxt.os << ")";
    }
};

std::ostream& operator<< (std::ostream& os, Expression& expr) {
    PrinterContext ctxt (os);
    PrinterPreVisitor preVisitor (ctxt);
    PrinterPostVisitor postVisitor (ctxt);

    collectingVisit(&expr, PrinterArgument(0, ""), preVisitor, postVisitor);

    return os;
}

}
