/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * Translation from parse trees to the internal SMT-LIB 2 abstract syntax. This
 * includes symbol resolution and type checking
 */

#ifndef SMTLIB_RESOLVER
#define SMTLIB_RESOLVER 1

#include <vector>
#include <map>
#include <exception>
#include <string>
#include <stack>
#include <cstdio>

#include "smtlib-absyn.H"
#include "theories/theory.H"

namespace Resolver {

using namespace SMTLibAbsyn;

/**
 * Superclass of all exceptions that can occur during resolution or type
 * checking
 */
class TranslationException : public std::exception {
private:
	std::string message;

public:
	TranslationException(const std::string& message);
	~TranslationException() throw ();

	void addMessage(const std::string& msg);

	const char* what() const throw();
};

/**
 * Exception raised in the Environment if a symbol to be looked up is not
 * declared
 */
class NoSuchSymbolException : public TranslationException {
public:
	NoSuchSymbolException(const std::string& message);
};

/**
 * Exception raised in the Environment if multiple definitions are available
 * for some function application
 */
class AmbiguityException : public TranslationException {
public:
	AmbiguityException(const std::string& message);
};

/**
 * To handle overloading (both parametric and ad-hoc), a set of functions with
 * the same name are grouped into "families".
 */
class FunctionFamily : public RefCounted {
public:
    MonoFunctionPtr
        lookup(const std::vector<SortPtr>& argumentSorts);

    virtual MonoFunctionPtr
		lookup(const std::vector<std::string>& indexes,
		       const std::vector<SortPtr>& argumentSorts) = 0;
    virtual MonoFunctionPtr
		lookup(const std::vector<std::string>& indexes,
		       const SortPtr& resSort,
		       const std::vector<SortPtr>& argumentSorts) = 0;

    ExpressionPtr m(const ExpressionPtr& arg0);
    ExpressionPtr m(const ExpressionPtr& arg0,
                    const ExpressionPtr& arg1);
    ExpressionPtr m(const ExpressionPtr& arg0,
                    const ExpressionPtr& arg1,
                    const ExpressionPtr& arg2);

    template <class I>
    ExpressionPtr m(I argBegin, I argEnd) {
        const std::vector<ExpressionPtr> arguments (argBegin, argEnd);
        return lookup(extractSorts(arguments))->m(arguments.begin(), arguments.end());
    }
};

typedef CountingPtr<FunctionFamily> FunctionFamilyPtr;

class MonoFunctionFamily : public FunctionFamily {
private:
    const MonoFunctionPtr function;

public:
    MonoFunctionFamily(const MonoFunctionPtr& function);

    MonoFunctionPtr
		lookup(const std::vector<std::string>& indexes,
		       const std::vector<SortPtr>& argumentSorts);
    MonoFunctionPtr
		lookup(const std::vector<std::string>& indexes,
		       const SortPtr& resSort,
		       const std::vector<SortPtr>& argumentSorts);
};

typedef std::multimap<std::string, FunctionFamilyPtr> FunMap;

/**
 * Similarly, we define families of sort constructors
 */
class SortCtorFamily : public RefCounted {
public:
	virtual SortCtorPtr lookup(const std::vector<std::string>& indexes) = 0;
};

class SimpleSortCtorFamily : public SortCtorFamily {
private:
	const SortCtorPtr sort;

public:
	SimpleSortCtorFamily(const SortCtorPtr& sort);

	SortCtorPtr lookup(const std::vector<std::string>& indexes);
};

typedef CountingPtr<SortCtorFamily> SortCtorFamilyPtr;
typedef std::multimap<std::string, SortCtorFamilyPtr> SortMap;
typedef std::map<std::string, SortPtr> SortSynonymMap;

enum LiteralKind { Numeral, Rational, Hexadecimal, Binary, String };

/**
 * Environment of function symbols, bound symbols, etc., when translating
 * parse trees into ASTs
 */
class Environment {
private:
    SortMap sortCtors;

    FunMap functions;

    SortSynonymMap sortSynonyms;

    std::vector<std::string> variableNames;
    std::vector<SortPtr> variableSorts;

    std::map<LiteralKind, SortPtr> literalSorts;

    VariablePtr lookupVariable(const std::string& name) const;

    void checkSortOverloading(const std::string& name);

public:

    void pushVariable(const std::string& variableName,
                      const SortPtr& variableSort);
    void popVariable();

    void addSortCtorFamily(const std::string& name,
                           const SortCtorFamilyPtr& sort,
                           bool allowOverloading);

    void addFunctionFamily(const std::string& name,
                           const FunctionFamilyPtr& function,
                           bool allowOverloading);

    void addLiteralSort(LiteralKind kind, const SortPtr& sort);

    void addSortSynoym(const std::string& name, const SortPtr& value);

    /**
	 * Lookup a function or variable with a given name and given argument
	 * (and possibly result) sorts. The returned object is either a MonoFunction
	 * or a Variable. In case no function with this signature exists,
	 * NoSuchSymbolException is raised. In case of more than one function with
	 * this signature exists, AmbiguityException is raised. In case the
	 * specified name denotes a variable, but at least one argument is given,
	 * VariableArgumentException is raised.
	 */
    RefCountedPtr lookup(const IndexedString& name,
                         const std::vector<SortPtr>& argumentSorts) const;
    RefCountedPtr lookup(const IndexedString& name,
                         const SortPtr& resSort,
                         const std::vector<SortPtr>& argumentSorts) const;

    /**
     * Lookup a sort constructor with a given name and indexes.
     * In case no constructor with this signature exists,
	 * NoSuchSymbolException is raised. In case of more than one constructor with
	 * this signature exists, AmbiguityException is raised.
     */
    SortCtorPtr lookupSortCtor(const IndexedString& name) const;

    /**
     * Lookup a sort synonym with a given name and indexes.
     * In case no synonym with this signature exists,
     * NoSuchSymbolException is raised.
     */
    SortPtr lookupSortSynonym(const IndexedString& name) const;

    /**
     * Lookup the sort of a literal. In case the sort of literals of this kind
     * is not defined, NoSuchSymbolException is raised.
     */
    SortPtr lookupLiteralSort(LiteralKind kind) const;
};

////////////////////////////////////////////////////////////////////////////////
//
// Datastructure for representing benchmarks
//
// TODO: this should better be immutable

struct Benchmark : RefCounted {
    std::string name;
    std::string logic;
    Theories::Theory* theory;

    std::vector<SortCtorPtr> extraSorts;
    std::vector<MonoFunctionPtr> extraFuns;

    std::vector<ExpressionPtr> assumptions;

    ExpressionPtr formula;
};

typedef CountingPtr<Benchmark> BenchmarkPtr;

BenchmarkPtr translateBenchmark(FILE *input);

}

#endif
