/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <algorithm>
#include <sstream>

#include "smtlib-resolver.H"
#include "Absyn.H"
#include "Printer.H"
#include "Parser.H"
#include "util.H"
#include "theories/empty.H"
#include "theories/sets.H"
#include "theories/lists.H"
#include "theories/maps.H"
#include "theories/relations.H"
#include "theories/lrm.H"
#include "multi-arity-ops.H"

using namespace std;

namespace Resolver {

TranslationException::TranslationException(const std::string& message)
    : message(message) {}

TranslationException::~TranslationException() throw () {}

void TranslationException::addMessage(const std::string& msg) {
    message = message + "\n" + msg;
}

const char* TranslationException::what() const throw() {
	return message.c_str();
}

NoSuchSymbolException::NoSuchSymbolException(const std::string& message)
	: TranslationException(message) {}

AmbiguityException::AmbiguityException(const std::string& message)
	: TranslationException(message) {}

////////////////////////////////////////////////////////////////////////////////

MonoFunctionPtr FunctionFamily::lookup(const std::vector<SortPtr>& argumentSorts) {
    return lookup(vector<string>(), argumentSorts);
}

ExpressionPtr FunctionFamily::m(const ExpressionPtr& arg0) {
    std::vector<SortPtr> argumentSorts;
    argumentSorts.push_back(arg0->sort());
    return lookup(argumentSorts)->m(arg0);
}

ExpressionPtr FunctionFamily::m(const ExpressionPtr& arg0,
                                const ExpressionPtr& arg1) {
    std::vector<SortPtr> argumentSorts;
    argumentSorts.push_back(arg0->sort());
    argumentSorts.push_back(arg1->sort());
    return lookup(argumentSorts)->m(arg0, arg1);
}
ExpressionPtr FunctionFamily::m(const ExpressionPtr& arg0,
                                const ExpressionPtr& arg1,
                                const ExpressionPtr& arg2) {
    std::vector<SortPtr> argumentSorts;
    argumentSorts.push_back(arg0->sort());
    argumentSorts.push_back(arg1->sort());
    argumentSorts.push_back(arg2->sort());
    return lookup(argumentSorts)->m(arg0, arg1, arg2);
}

////////////////////////////////////////////////////////////////////////////////

MonoFunctionFamily::MonoFunctionFamily(const MonoFunctionPtr& function)
    : function(function) {}

MonoFunctionPtr MonoFunctionFamily::lookup(const std::vector<std::string>& indexes,
                            		       const std::vector<SortPtr>& argumentSorts) {
    if (function->name.indexes == indexes &&
    	equalVectors(function->argumentSorts, argumentSorts))
        return function;
    throw NoSuchSymbolException("No such function is declared");
}

MonoFunctionPtr MonoFunctionFamily::lookup(const std::vector<std::string>& indexes,
                            		       const SortPtr& resSort,
                                           const std::vector<SortPtr>& argumentSorts) {
    if (function->name.indexes == indexes &&
        equalVectors(function->argumentSorts, argumentSorts) &&
        *resSort == *function->resultSort)
        return function;
    throw NoSuchSymbolException("No such function is declared");
}

////////////////////////////////////////////////////////////////////////////////

SimpleSortCtorFamily::SimpleSortCtorFamily(const SortCtorPtr& sort)
	: sort(sort) {}

SortCtorPtr SimpleSortCtorFamily::lookup(const std::vector<std::string>& indexes) {
	if (indexes.size() > 0)
		throw NoSuchSymbolException("No such sort is declared");
	return sort;
}

////////////////////////////////////////////////////////////////////////////////

void Environment::pushVariable(const std::string& variableName,
                               const SortPtr& variableSort) {
    variableNames.push_back(variableName);
    variableSorts.push_back(variableSort);
}

void Environment::popVariable() {
    assert(!variableNames.empty());
    variableNames.pop_back();
    variableSorts.pop_back();
}

void Environment::checkSortOverloading(const std::string& name) {
    pair<SortMap::iterator, SortMap::iterator> familyRange =
        sortCtors.equal_range(name);
    if (familyRange.first != familyRange.second)
        throw TranslationException(
            "The sort " + name + " has already been declared");
    if (sortSynonyms.find(name) != sortSynonyms.end())
        throw TranslationException(
            "The identifier " + name +
            " has already been declared as sort synonym");
}

void Environment::addSortCtorFamily(const std::string& name,
                                    const SortCtorFamilyPtr& sort,
                                    bool allowOverloading) {
    if (!allowOverloading)
        // check that no sorts with the given name are registered
        checkSortOverloading(name);

    sortCtors.insert(pair<string, SortCtorFamilyPtr>(name, sort));
}

void Environment::addFunctionFamily(const std::string& name,
                                    const FunctionFamilyPtr& function,
                                    bool allowOverloading) {
    if (!allowOverloading) {
        // check that no functions with the given name are registered
        pair<FunMap::iterator, FunMap::iterator> familyRange =
            functions.equal_range(name);
        if (familyRange.first != familyRange.second)
            throw TranslationException(
                "The function " + name + " has already been declared");
    }

    functions.insert(FunMap::value_type(name, function));
}

void Environment::addLiteralSort(LiteralKind kind, const SortPtr& sort) {
    assert(literalSorts.find(kind) == literalSorts.end());
    literalSorts.insert(pair<LiteralKind, SortPtr>(kind, sort));
}

void Environment::addSortSynoym(const std::string& name, const SortPtr& value) {
    // check that no sorts with the given name are registered
    checkSortOverloading(name);

    sortSynonyms[name] = value;
}

VariablePtr Environment::lookupVariable(const std::string& name) const {
    vector<string>::const_reverse_iterator pos =
        find(variableNames.rbegin(), variableNames.rend(), name);
    if (pos == variableNames.rend())
        return 0;

    int index = pos - variableNames.rbegin();
    SortPtr sort = variableSorts[variableSorts.size() - index - 1];

    return v(sort, index);
}

RefCountedPtr Environment::lookup(const IndexedString& name,
                                  const std::vector<SortPtr>& argumentSorts) const {
    return lookup(name, 0, argumentSorts);
}

RefCountedPtr Environment::lookup(const IndexedString& name,
                                  const SortPtr& resSort,
                                  const std::vector<SortPtr>& argumentSorts) const {
	if (name.indexes.empty()) {
		VariablePtr variable = lookupVariable(name.body);
		if (variable) {
			if (!argumentSorts.empty())
				throw TranslationException(
				    "Variable " + name.body + " cannot receive arguments");
			return variable;
		}
	}

    pair<FunMap::const_iterator, FunMap::const_iterator> familyRange =
        functions.equal_range(name.body);

    MonoFunctionPtr function;

    for (FunMap::const_iterator it = familyRange.first;
         it != familyRange.second;
         ++it) {
        try {
            MonoFunctionPtr newFunction =
                resSort ? it->second->lookup(name.indexes, resSort, argumentSorts)
                        : it->second->lookup(name.indexes, argumentSorts);
            if (newFunction) {
                if (function)
                    throw AmbiguityException(
                        "More than one declaration exists for " + name.toString());
                function = newFunction;
            }
        } catch (NoSuchSymbolException&) {
            // nothing, we just continue with the next definition of functions
            // with this name
        }
    }

    if (function)
        return function;

    stringstream ss;
    ss << "The symbol " << name;
    if (!argumentSorts.empty()) {
        ss << " with argument sorts (";
        for_each(argumentSorts.begin(), argumentSorts.end(),
                 print_deref<SortPtr>(ss, " "));
        ss << ")";
    }
    if (resSort) {
        if (argumentSorts.empty())
            ss << " of ";
        else
            ss << " and result ";
        ss << "sort " << *resSort;
    }
    ss << " is not declared";

    throw NoSuchSymbolException(ss.str());
}

SortCtorPtr Environment::lookupSortCtor(const IndexedString& name) const {
    pair<SortMap::const_iterator, SortMap::const_iterator> familyRange =
    	sortCtors.equal_range(name.body);

    SortCtorPtr sortCtor;

    for (SortMap::const_iterator it = familyRange.first;
         it != familyRange.second;
         ++it) {
    	try {
    		SortCtorPtr newSortCtor = it->second->lookup(name.indexes);
    		if (newSortCtor) {
    			if (sortCtor)
                    throw AmbiguityException(
                        "More than one declaration exists for the sort " + name.toString());
    			sortCtor = newSortCtor;
    		}
    	} catch (NoSuchSymbolException&) {
            // nothing, we just continue with the next definition of constructors
            // with this name
    	}
    }

    if (sortCtor)
    	return sortCtor;

    throw NoSuchSymbolException("The sort " + name.toString() + " is not declared");
}

SortPtr Environment::lookupSortSynonym(const IndexedString& name) const {
    if (!name.indexes.empty())
        throw NoSuchSymbolException(
            "The sort synonym " + name.toString() + " is not declared");

    const SortSynonymMap::const_iterator pos = sortSynonyms.find(name.body);
    if (pos == sortSynonyms.end())
        throw NoSuchSymbolException(
            "The sort synonym " + name.toString() + " is not declared");

    return pos->second;
}

SortPtr Environment::lookupLiteralSort(LiteralKind kind) const {
    map<LiteralKind, SortPtr>::const_iterator it = literalSorts.find(kind);
    if (it == literalSorts.end()) {
        stringstream ss;
        ss << "Literals of the kind " << kind << " are not declared";
        throw NoSuchSymbolException(ss.str());
    }
    return it->second;
}

////////////////////////////////////////////////////////////////////////////////
//
// Local class to translate parse trees to the internal SMT-LIB ASTs. This is
// iteratively and in depth-first order (and not recursively), using two stacks
// for the nodes still to be visited, and the results of visiting nodes.

class ParseTreeTranslator {
private:
    Environment env;

	stack<const Term*> toVisit;
	stack<ExpressionPtr> results;

	/**
	 * We generate a couple of distinguished terms that are used to remember
	 * positions on the stack
	 */
	class PseudoTerm : public Term {
	    void accept(Visitor* v) {
	        assert(false);
	    }
	    Term* clone() const {
	        assert(false);
            return 0; // make the compiler happy
	    }
	};

	/**
	 * Term that is used to remember places on the stack
	 * where Term nodes can be reassembled to an Expression
	 */
    static const PseudoTerm *ASSEMBLE_EXPRESSION;

    /**
     * Term that is used to remember places on the stack
     * where the variables in a let-binding can be bound. This is necessary
     * because only the body-expression is the scope of a let-binder, but the
     * let-expression can have many more sub-expressions.
     */
    static const PseudoTerm *LET_BIND_VARIABLES;

	IndexedString translateIdentifier(const ::Identifier *id) {
		if (const SymbolIdent *ident = dynamic_cast<const SymbolIdent*>(id)) {
			return IndexedString(ident->symbol_);
		} else if (const IndexIdent *ident = dynamic_cast<const IndexIdent*>(id)) {
			vector<string> indexes;
			ListIndexC *rawIndexes = ident->listindexc_;
			while (rawIndexes != 0) {
				indexes.push_back(dynamic_cast<Index*>(rawIndexes->indexc_)->numeral_);
				rawIndexes = rawIndexes->listindexc_;
			}

			return IndexedString(ident->symbol_, indexes);
		}
		assert(false);
	}

	void throwSortArityException(const SortCtorPtr& sortCtor) {
		stringstream ss;
		ss << "The sort constructor " << sortCtor->name <<  " needs to receive "
		   << sortCtor->arity << " arguments";
		throw TranslationException(ss.str());
	}

	SortPtr translateSort(const ::Sort *sort) {
		if (const IdentSort *identSort =
		        dynamic_cast<const IdentSort*>(sort)) {

			const IndexedString name = translateIdentifier(identSort->identifier_);
			try {
			    return env.lookupSortSynonym(name);
			} catch (NoSuchSymbolException&) {}

			const SortCtorPtr sortCtor = env.lookupSortCtor(name);

			if (sortCtor->arity > 0)
				throwSortArityException(sortCtor);

			return sortCtor->m();

		} else if (const CompositeSort *compositeSort =
		               dynamic_cast<const CompositeSort*>(sort)) {

			const IndexedString name = translateIdentifier(compositeSort->identifier_);
			const SortCtorPtr sortCtor = env.lookupSortCtor(name);

			vector<SortPtr> arguments;
			ListSort *rawArguments = compositeSort->listsort_;
			while (rawArguments != 0) {
				arguments.push_back(translateSort(rawArguments->sort_));
				rawArguments = rawArguments->listsort_;
			}

			if (arguments.size() != sortCtor->arity)
				throwSortArityException(sortCtor);

			return new SMTLibAbsyn::Sort(sortCtor, arguments.begin(), arguments.end());

		} else {
		    assert(false);
		}
	}

	ExpressionPtr translateSymbolApp(const SymbolRef* symbol,
	                                 const vector<ExpressionPtr>& arguments) {
	    vector<SortPtr> argumentSorts = extractSorts(arguments);

	    RefCountedPtr appliedSymbol;
	    if (const IdentifierRef* identifierRef =
	            dynamic_cast<const IdentifierRef*>(symbol)) {
	        const IndexedString id = translateIdentifier(identifierRef->identifier_);
	        appliedSymbol = env.lookup(id, argumentSorts);
	    } else if (const CastIdentifierRef* castIdentifierRef =
	                   dynamic_cast<const CastIdentifierRef*>(symbol)) {
            const IndexedString id = translateIdentifier(castIdentifierRef->identifier_);
            const SortPtr resultSort = translateSort(castIdentifierRef->sort_);
            appliedSymbol = env.lookup(id, resultSort, argumentSorts);

            if (!(MonoFunctionPtr)appliedSymbol) {
                PrintAbsyn p;
                stringstream ss;
                ss << "Casts can only be applied to function symbols, not to "
                   << p.print((SymbolRef*)symbol);
                throw TranslationException(ss.str());
            }
	    }

	    assert(appliedSymbol);

	    if (const VariablePtr var = (VariablePtr)appliedSymbol) {
	        return var;
	    } else if (const ExpandableMonoFunctionPtr function =
	                   (ExpandableMonoFunctionPtr)appliedSymbol) {
	        return function->expand(arguments);
	    } else if (const MonoFunctionPtr function =
	                   (MonoFunctionPtr)appliedSymbol) {
            return new FunctionApp(function, arguments.begin(), arguments.end());
	    }

	    assert(false);
	}

	/**
	 * Push the sub-terms of the given term on the toVisit stack
	 */
	void pushSubTerms(const Term *term) {
	    // we first push the original term, so that we later know what we have
	    // to assemble
	    toVisit.push(term);
	    toVisit.push(ASSEMBLE_EXPRESSION);

	    if (dynamic_cast<const NullaryTerm*>(term) != 0) {
	        // nothing, no subterms
	    } else if (const FunctionTerm *functionTerm =
	                   dynamic_cast<const FunctionTerm*>(term)) {
            ListTerm *subterms = functionTerm->listterm_;
            while (subterms != 0) {
                toVisit.push(subterms->term_);
                subterms = subterms->listterm_;
            }
        } else if (dynamic_cast<const ConstantTerm*>(term) != 0) {
            // nothing, no subterms
        } else if (const LetTerm *letTerm = dynamic_cast<const LetTerm*>(term)) {
            // the last subterm to be processed is the body (because to bind
            // the variables for the body, we first need to infer the types of
            // the bound variables)
            toVisit.push(letTerm->term_);
            toVisit.push(letTerm);
            toVisit.push(LET_BIND_VARIABLES);

            // then push the terms denoted by the variables
            ListBindingC* bindings = letTerm->listbindingc_;
            while (bindings != 0) {
                toVisit.push(dynamic_cast<Binding*>(bindings->bindingc_)->term_);
                bindings = bindings->listbindingc_;
            }
        } else if (const QuantifierTerm *quantifierTerm =
                       dynamic_cast<const QuantifierTerm*>(term)) {
            toVisit.push(quantifierTerm->term_);

            ListSortedVariableC* sortedVars = quantifierTerm->listsortedvariablec_;
            while (sortedVars != 0) {
                SortedVariable* var =
                    dynamic_cast<SortedVariable*>(sortedVars->sortedvariablec_);
                env.pushVariable(var->symbol_, translateSort(var->sort_));
                sortedVars = sortedVars->listsortedvariablec_;
            }
        } else {
            assert(false);
        }
	}

	/**
	 * Read results from the result stack and assemble an expression. The term
	 * given as argument is the original term
	 */
	ExpressionPtr assembleExpression(const Term* term) {
        if (const NullaryTerm* nullaryTerm =
                dynamic_cast<const NullaryTerm*>(term)) {

            return translateSymbolApp(nullaryTerm->symbolref_, vector<ExpressionPtr>());

        } else if (const FunctionTerm *functionTerm =
                       dynamic_cast<const FunctionTerm*>(term)) {

            vector<ExpressionPtr> arguments;
            ListTerm *subterms = functionTerm->listterm_;
            while (subterms != 0) {
                arguments.push_back(results.top());
                results.pop();
                subterms = subterms->listterm_;
            }
            return translateSymbolApp(functionTerm->symbolref_, arguments);

        } else if (const ConstantTerm* constantTerm =
                       dynamic_cast<const ConstantTerm*>(term)) {

            const SpecConstant* constant = constantTerm->specconstant_;
            LiteralKind kind;
            string value;

            if (const NumConstant *c =
                    dynamic_cast<const NumConstant*>(constant)) {
                kind = Numeral;
                value = c->numeral_;
            } else if (const RatConstant *c =
                           dynamic_cast<const RatConstant*>(constant)) {
                kind = Rational;
                value = c->rational_;
            } else if (const HexConstant *c =
                           dynamic_cast<const HexConstant*>(constant)) {
                kind = Hexadecimal;
                value = c->hexadecimal_;
            } else if (const BinConstant *c =
                           dynamic_cast<const BinConstant*>(constant)) {
                kind = Binary;
                value = c->binary_;
            } else if (const StringConstant *c =
                           dynamic_cast<const StringConstant*>(constant)) {
                kind = String;
                value = c->string_;
            } else {
                assert(false);
            }
            return new Literal(env.lookupLiteralSort(kind), value);

        } else if (const LetTerm *letTerm = dynamic_cast<const LetTerm*>(term)) {

            const ExpressionPtr body = results.top();
            results.pop();

            vector<string> variableNames;
            vector<SortPtr> variableSorts;
            vector<ExpressionPtr> boundTerms;

            ListBindingC* bindings = letTerm->listbindingc_;
            while (bindings != 0) {
                const ExpressionPtr boundTerm = results.top();
                results.pop();

                Binding* binding = dynamic_cast<Binding*>(bindings->bindingc_);

                boundTerms.push_back(boundTerm);
                variableNames.push_back(binding->symbol_);
                variableSorts.push_back(boundTerm->sort());

                env.popVariable();

                bindings = bindings->listbindingc_;
            }

            return new LetBinder(variableNames.begin(), variableNames.end(),
                                 variableSorts.begin(), variableSorts.end(),
                                 boundTerms.begin(), boundTerms.end(), body);

        } else if (const QuantifierTerm *quantifierTerm =
                       dynamic_cast<const QuantifierTerm*>(term)) {

            const ExpressionPtr body = results.top();
            results.pop();

            if (!(*body->sort() == *SMTLibAbsyn::Sort::Bool))
                throw TranslationException (
                    "The body of a quantifier has to be a boolean expression");

            vector<string> variableNames;
            vector<SortPtr> variableSorts;

            ListSortedVariableC* sortedVars = quantifierTerm->listsortedvariablec_;
            while (sortedVars != 0) {
                SortedVariable* var =
                    dynamic_cast<SortedVariable*>(sortedVars->sortedvariablec_);

                variableNames.push_back(var->symbol_);
                variableSorts.push_back(translateSort(var->sort_));

                sortedVars = sortedVars->listsortedvariablec_;

                env.popVariable();
            }

            SMTLibAbsyn::Quantifier quantifier;
            if (dynamic_cast<AllQuantifier*>(quantifierTerm->quantifier_) != 0)
                quantifier = Forall;
            else if (dynamic_cast<ExQuantifier*>(quantifierTerm->quantifier_) != 0)
                quantifier = Exists;
            else
                assert(false);

            return new QuantifiedExpression(quantifier,
                                            variableNames.begin(), variableNames.end(),
                                            variableSorts.begin(), variableSorts.end(),
                                            body);

        } else {
            assert(false);
        }
	}

	/**
	 * Pop a term from the toVisit stack and find out what needs to be done
	 */
	void processSubTerm() {
	    const Term* topTerm = toVisit.top();
	    toVisit.pop();
	    if (topTerm == ASSEMBLE_EXPRESSION) {
	        const Term* originalTerm = toVisit.top();
	        toVisit.pop();
	        try {
	            results.push(assembleExpression(originalTerm));
	        } catch (TranslationException &e) {
                e.addMessage(exceptionFrame(originalTerm));
                throw e;
	        }
	    } else if (topTerm == LET_BIND_VARIABLES) {
            const Term* originalTerm = toVisit.top();
            toVisit.pop();
	        stack<ExpressionPtr> topResults;

            assert(dynamic_cast<const LetTerm*>(originalTerm));

	        // we need to peek at the top elements on the results stack, to
	        // find out about the sorts of the bound variables
	        ListBindingC* bindings = ((const LetTerm*)originalTerm)->listbindingc_;
            while (bindings != 0) {
                ExpressionPtr boundTerm = results.top();
                results.pop();
                topResults.push(boundTerm);

                Binding* binding = dynamic_cast<Binding*>(bindings->bindingc_);
                env.pushVariable(binding->symbol_, boundTerm->sort());

                bindings = bindings->listbindingc_;
            }

            // put the results back on the results stack, we need them later
            // to reassemble the expression
            while (!topResults.empty()) {
                results.push(topResults.top());
                topResults.pop();
            }
	    } else {
	        pushSubTerms(topTerm);
	    }
	}

    /**
     * Extract the :logic attribute of a benchmark and instantiate the right
     * theory
     */
    void registerTheory(BenchmarkPtr benchmark, const LogicAttribute* logicAttr) {
        benchmark->logic = logicAttr->symbol_;
        if (benchmark->logic == "FiniteSetWithCard")
            benchmark->theory = Theories::Sets::create();
        else if (benchmark->logic == "ListWithLength")
            benchmark->theory = Theories::Lists::create();
        else if (benchmark->logic == "FiniteMap")
            benchmark->theory = Theories::Maps::create();
        else if (benchmark->logic == "FiniteRelationWithCard")
            benchmark->theory = Theories::Relations::create();
        else if (benchmark->logic == "LRM")
            benchmark->theory = Theories::LRM::create();
        else
            throw TranslationException ("Unsupported :logic");

        benchmark->theory->registerSymbols(env);
    }

    Theories::Empty getEmpty(BenchmarkPtr benchmark) {
    	return ((Theories::TheoryHelpers*)benchmark->theory)->getEmpty();
    }

    /**
     * Add the :extrasorts defined in a benchmark to the environment and the
     * internal benchmark representation
     */
    void registerSorts(BenchmarkPtr benchmark,
                       const ExtraSortsAttribute* sortsAttr) {
        ListSortDeclC* sorts = sortsAttr->listsortdeclc_;
        while (sorts != 0) {
            if (const SortDecl* sortDecl =
                dynamic_cast<SortDecl*>(sorts->sortdeclc_)) {

                IndexedString name = translateIdentifier(sortDecl->identifier_);

                if (!name.indexes.empty())
                    throw TranslationException (
                        "Extra sort " + name.body + " is not allowed to be indexed");

                stringstream ss (sortDecl->numeral_);
                unsigned int arity;
                ss >> arity;

                SortCtorPtr sortCtor = new SortCtor(name, arity);

                env.addSortCtorFamily(name.body,
                                      new SimpleSortCtorFamily(sortCtor),
                                      false);
                benchmark->extraSorts.push_back(sortCtor);

                // TODO: take care of the annotations

            } else if (const SortSynonymDecl* sortDecl =
                           dynamic_cast<SortSynonymDecl*>(sorts->sortdeclc_)) {

                IndexedString name = translateIdentifier(sortDecl->identifier_);

                if (!name.indexes.empty())
                    throw TranslationException (
                        "Extra sort " + name.body + " is not allowed to be indexed");

                const SortPtr value = translateSort(sortDecl->sort_);
                env.addSortSynoym(name.body, value);

                // TODO: take care of the annotations

            } else {
                assert(false);
            }

            sorts = sorts->listsortdeclc_;
        }
    }

    /**
     * Add the :extrafuns defined in a benchmark to the environment and the
     * internal benchmark representation
     */
    void registerFunctions(BenchmarkPtr benchmark,
                           const ExtraFunsAttribute* funsAttr) {
        ListFunctionDeclC* funs = funsAttr->listfunctiondeclc_;
        while (funs != 0) {

            if (const FunctionDecl* functionDecl =
                    dynamic_cast<FunctionDecl*>(funs->functiondeclc_)) {

                IndexedString name = translateIdentifier(functionDecl->identifier_);

                if (!name.indexes.empty())
                    throw TranslationException (
                        "Extra function " + name.body + " is not allowed to be indexed");

                ListSort* sorts = functionDecl->listsort_;

                vector<SortPtr> argumentSorts;
                while (sorts != 0) {
                    argumentSorts.push_back(translateSort(sorts->sort_));
                    sorts = sorts->listsort_;
                }

                SortPtr resSort = argumentSorts.back();
                argumentSorts.pop_back();

                MonoFunctionPtr function =
                    new MonoFunction(name, resSort,
                                     argumentSorts.begin(), argumentSorts.end());

                env.addFunctionFamily(name.body,
                                      new MonoFunctionFamily(function),
                                      false);
                benchmark->extraFuns.push_back(function);

                // TODO: take care of the annotations

            } else if (const MacroDecl* macroDecl =
                       dynamic_cast<MacroDecl*>(funs->functiondeclc_)) {

                IndexedString name = translateIdentifier(macroDecl->identifier_);

                if (!name.indexes.empty())
                    throw TranslationException (
                        "Macro " + name.body + " is not allowed to be indexed");

                // Declare the bound variables, ensure that the bound variables
                // are all distinct
                unsigned int boundVarNum = 0;
                set<string> boundVars;

                vector<string> variableNames;
                vector<SortPtr> variableSorts;

                ListSortedVariableC* vars = macroDecl->listsortedvariablec_;
                while (vars != 0) {
                	boundVarNum = boundVarNum + 1;
                	const SortedVariable* var =
                			(SortedVariable*)vars->sortedvariablec_;
                	boundVars.insert(var->symbol_);

                	const SortPtr varSort = translateSort(var->sort_);
                	env.pushVariable(var->symbol_, varSort);

                	variableNames.push_back(var->symbol_);
                	variableSorts.push_back(varSort);

                	vars = vars->listsortedvariablec_;
                }

                if (boundVars.size() != boundVarNum)
                	throw TranslationException(
                	    "Parameter names are not distinct in definition of macro " +
                	    name.body);

                const ExpressionPtr functionBody = translateTerm(macroDecl->term_);

                for (unsigned int i = 0; i < boundVarNum; ++i)
                	env.popVariable();

                const MonoFunctionPtr function =
                    new MonoFunction (name, functionBody->sort(), variableSorts);
                env.addFunctionFamily(name.body,
                                      new MonoFunctionFamily(function), false);
                benchmark->extraFuns.push_back(function);

                // Generate an axiom that defines the macro
                vector<ExpressionPtr> arguments;
                for (unsigned int i = 0; i < variableSorts.size(); ++i)
                	arguments.push_back(v(variableSorts[i],
                	                      variableSorts.size() - i - 1));
                const ExpressionPtr lhs =
                	function->m(arguments.begin(), arguments.end());
                const ExpressionPtr eq =
                    getEmpty(benchmark).polyfun_equal->m(lhs, functionBody);

                const ExpressionPtr axiom =
                	new QuantifiedExpression(Forall,
                	                         variableNames.begin(), variableNames.end(),
                	                         variableSorts.begin(), variableSorts.end(),
                	                         eq);

                benchmark->assumptions.push_back(axiom);

                // TODO: take care of the annotations

            } else {
                assert(false);
            }

            funs = funs->listfunctiondeclc_;
        }
    }

    /**
     * Add the :assumptions defined in a benchmark to the internal benchmark
     * representation
     */
    void translateAssumption(BenchmarkPtr benchmark,
                             const AssumptionAttribute* assumptionAttr) {
        const ExpressionPtr assumption = translateTerm(assumptionAttr->term_);

        if (!(*assumption->sort() == *SMTLibAbsyn::Sort::Bool))
            throw TranslationException (
                     ":assumption has to be a boolean expression");
            benchmark->assumptions.push_back(assumption);
    }

    /**
     * Add the :formula defined in a benchmark to the internal benchmark
     * representation (it is assumed that there is only one :formula attribute)
     */
    void translateFormula(BenchmarkPtr benchmark,
                          const FormulaAttribute* formulaAttr) {
        if (benchmark->formula)
            throw TranslationException (
                        "Benchmark must not contain more than one :formula");

        benchmark->formula = translateTerm(formulaAttr->term_);

        if (!(*benchmark->formula->sort() == *SMTLibAbsyn::Sort::Bool))
            throw TranslationException (":formula has to be a boolean expression");
    }

    void checkLogicIsSet(bool theorySet, const string& attribute) {
        if (!theorySet)
            throw TranslationException (
               "Attribute " + attribute + " must not be used before :logic");
    }

    ExpressionPtr translateTerm(Term* root) {
        toVisit.push(root);
        while (!toVisit.empty())
            processSubTerm();

        assert(results.size() == 1);
        const ExpressionPtr res = results.top();
        results.pop();

        return res;
    }

    string exceptionFrame(const Visitable *v) {
        PrintAbsyn p;
        stringstream ss;
        ss << "  in " << p.print((Visitable*)v);
        return ss.str();
    }

public:
	void translateBenchmark(BenchmarkPtr benchmark, ListAttribute* attributes) {
        ListAttribute* attr = attributes;
        bool theorySet = false;
        while (attr != 0) {
            try {
                if (const LogicAttribute* logicAttr =
                    dynamic_cast<LogicAttribute*>(attr->attribute_)) {
                    if (theorySet)
                        throw TranslationException (
                             "Attribute :logic may only be given once");

                    registerTheory(benchmark, logicAttr);
                    theorySet = true;
                } else if (const ExtraSortsAttribute* sortsAttr =
                           dynamic_cast<ExtraSortsAttribute*>(attr->attribute_)) {
                    checkLogicIsSet(theorySet, ":extrasorts");
                    registerSorts(benchmark, sortsAttr);
                } else if (const ExtraFunsAttribute* funsAttr =
                           dynamic_cast<ExtraFunsAttribute*>(attr->attribute_)) {
                    checkLogicIsSet(theorySet, ":extrafuns");
                    registerFunctions(benchmark, funsAttr);
                } else if (const AssumptionAttribute* assumptionAttr =
                           dynamic_cast<AssumptionAttribute*>(attr->attribute_)) {
                    checkLogicIsSet(theorySet, ":assumption");
                    translateAssumption(benchmark, assumptionAttr);
                } else if (const FormulaAttribute* formulaAttr =
                           dynamic_cast<FormulaAttribute*>(attr->attribute_)) {
                    checkLogicIsSet(theorySet, ":formula");
                    translateFormula(benchmark, formulaAttr);
                } else {
                    PrintAbsyn p;
                    cerr << "Ignoring " << p.print(attr->attribute_) << endl;
                }
            } catch (TranslationException &e) {
                e.addMessage(exceptionFrame(attr->attribute_));
                throw e;
            }

            attr = attr->listattribute_;
        }

        if (!benchmark->formula)
        	throw TranslationException ("Benchmark must contain a :formula");
	}
};

const ParseTreeTranslator::PseudoTerm *ParseTreeTranslator::ASSEMBLE_EXPRESSION =
    new ParseTreeTranslator::PseudoTerm;

const ParseTreeTranslator::PseudoTerm *ParseTreeTranslator::LET_BIND_VARIABLES =
    new ParseTreeTranslator::PseudoTerm;

////////////////////////////////////////////////////////////////////////////////

BenchmarkPtr translateBenchmark(FILE *input) {
    BenchmarkC *parse_tree = pBenchmarkC(input);

    if (!parse_tree)
        throw TranslationException("Parsing failed");

//    ShowAbsyn *s = new ShowAbsyn();
//    printf("%s\n\n", s->show(parse_tree));

    ::Benchmark* rawBenchmark = dynamic_cast< ::Benchmark*>(parse_tree);
    ListAttribute* attributes = rawBenchmark->listattribute_;

    BenchmarkPtr res = new Benchmark;
    res->name = rawBenchmark->symbol_;

    ParseTreeTranslator translator;
    translator.translateBenchmark(res, attributes);

    return res;
}

}

