/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "substitutions.H"
#include "util.H"

namespace Substitutions {

using namespace Visitors;
using namespace SMTLibAbsyn;
using namespace std;

struct VariableShiftContext {
    unsigned int offset;
    int shift;

    VariableShiftContext() : offset(0), shift(0) {}
    VariableShiftContext(int offset, int shift) : offset(offset), shift(shift) {
        assert(offset + shift >= 0);
    }
};

class VariableShiftPreVisitor
    : public DefaultExpressionVisitor<VariableShiftContext, int, PreVisitResultPtr> {
public:
    void defaultAction(const ExpressionPtr& node) {
        res = keepArg();
    }

    void visitLetBinder           (const CountingPtr<LetBinder>& node) {
        vector<VariableShiftContext> arguments (node->boundTerms.size(), arg0);
        arguments.push_back(VariableShiftContext(arg0.offset + node->boundTerms.size(),
                                                 arg0.shift));
        res = subArgs(arguments);
    }

    void visitQuantifiedExpression(const CountingPtr<QuantifiedExpression>& node) {
        res = uniSubArgs(VariableShiftContext(arg0.offset + node->variableNames.size(),
                                              arg0.shift));
    }
};

class VariableShiftPostVisitor
    : public DefaultExpressionVisitor<VariableShiftContext,
                                      const std::vector<ExpressionPtr>*,
                                      ExpressionPtr> {
public:
    void defaultAction(const ExpressionPtr& node) {
        res = node->update(*arg1);
    }

    void visitVariable(const CountingPtr<Variable>& node) {
        if (node->index >= arg0.offset && arg0.shift != 0)
            res = v(node->sort(), node->index + arg0.shift);
        else
            res = node;
    }
};

ExpressionPtr shiftVariables(const ExpressionPtr& expr,
                             unsigned int offset, int shift) {
    VariableShiftPreVisitor preVisitor;
    VariableShiftPostVisitor postVisitor;
    return collectingVisit(expr, VariableShiftContext(offset, shift),
                           preVisitor, postVisitor);
}

////////////////////////////////////////////////////////////////////////////////
// Eliminate let-expressions whose sort is not boolean (such expressions are
// not supported in SMT-LIB 1)

typedef SinglyLinkedList<ExpressionPtr> ExpressionList;
typedef CountingPtr<ExpressionList> ExpressionListPtr;

ExpressionPtr eliminateLet(const ExpressionPtr& expr,
                           const ExpressionListPtr& subst);

class LetEliminationPreVisitor
    : public DefaultExpressionVisitor<ExpressionListPtr, int, PreVisitResultPtr> {

private:
    ExpressionListPtr shiftVariablesInList(ExpressionListPtr exprs, int shift) {
        if (exprs)
            return new ExpressionList(shiftVariables(exprs->value, 0, shift),
                                      shiftVariablesInList(exprs->next, shift));
        else
            return 0;
    }

    ExpressionListPtr prependVariables(ExpressionListPtr exprs,
                                       const vector<SortPtr>& sorts) {
        ExpressionListPtr res = exprs;
        int i = sorts.size() - 1;
        for (vector<SortPtr>::const_iterator it = sorts.begin();
             it != sorts.end();
             ++it, --i)
            res = new ExpressionList(v(*it, i), res);
        return res;
    }

public:
    void defaultAction(const ExpressionPtr& node) {
        res = keepArg();
    }

    void visitLetBinder           (const CountingPtr<LetBinder>& node) {
        if (*node->sort() == *Sort::Bool) {
            // just descend and apply the substitutions underneath the let
            vector<ExpressionListPtr> arguments (node->variableSorts.size(), arg0);
            arguments.push_back(prependVariables(
                                   shiftVariablesInList(arg0, node->variableSorts.size()),
                                   node->variableSorts));
            res = subArgs(arguments);
        } else {
            // eliminate the let-expression
            ExpressionListPtr context = arg0;
            for (vector<ExpressionPtr>::const_iterator it = node->boundTerms.begin();
                 it != node->boundTerms.end();
                 ++it)
                context = new ExpressionList(eliminateLet(*it, arg0), context);
            res = tryAgain(node->body, context);
        }
    }

    void visitQuantifiedExpression(const CountingPtr<QuantifiedExpression>& node) {
        res = uniSubArgs(prependVariables(
                          shiftVariablesInList(arg0, node->variableSorts.size()),
                          node->variableSorts));
    }
};

class LetEliminationPostVisitor
    : public DefaultExpressionVisitor<ExpressionListPtr,
                                      const std::vector<ExpressionPtr>*,
                                      ExpressionPtr> {
    void defaultAction(const ExpressionPtr& node) {
        res = node->update(*arg1);
    }

    void visitVariable(const CountingPtr<Variable>& node) {
        unsigned int substLen = listLength(arg0);
        if (substLen == 0)
            res = node;
        else if (node->index >= substLen)
            res = v(node->sort(), node->index - substLen);
        else
            res = arg0->drop(node->index)->value;
    }
};

ExpressionPtr eliminateLet(const ExpressionPtr& expr,
                           const ExpressionListPtr& subst) {
    LetEliminationPreVisitor preVisitor;
    LetEliminationPostVisitor postVisitor;
    return collectingVisit(expr, subst, preVisitor, postVisitor);
}

ExpressionPtr eliminateLet(const ExpressionPtr& expr) {
    return eliminateLet(expr, 0);
}

void eliminateLet(const Resolver::BenchmarkPtr& benchmark) {
    for (vector<ExpressionPtr>::iterator it = benchmark->assumptions.begin();
         it != benchmark->assumptions.end();
         ++it)
        *it = eliminateLet(*it);
    benchmark->formula = eliminateLet(benchmark->formula);
}

}
