/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "lists.H"
#include "../util.H"
#include "function-visitor.H"

namespace Theories {

using namespace SMTLibAbsyn;
using namespace std;

PolyFunctionFamilyPtr createCons(const SortCtorPtr& sort_List) {
    const SortVariablePtr alpha = new SortVariable;
    const SortPtr List_alpha = sort_List->m(alpha->m());
    return new PolyFunctionFamily("cons", List_alpha,
                                  to_vector(alpha->m(), List_alpha));
}

PolyFunctionFamilyPtr createHead(const SortCtorPtr& sort_List) {
    const SortVariablePtr alpha = new SortVariable;
    return new PolyFunctionFamily("head", alpha->m(),
                                  to_vector(sort_List->m(alpha->m())));
}

PolyFunctionFamilyPtr createTail(const SortCtorPtr& sort_List) {
    const SortVariablePtr alpha = new SortVariable;
    const SortPtr List_alpha = sort_List->m(alpha->m());
    return new PolyFunctionFamily("tail", List_alpha, to_vector(List_alpha));
}

PolyFunctionFamilyPtr createAppend(const SortCtorPtr& sort_List) {
    const SortVariablePtr alpha = new SortVariable;
    const SortPtr List_alpha = sort_List->m(alpha->m());
    return new PolyFunctionFamily("append", List_alpha,
                                  to_vector(List_alpha, List_alpha));
}

PolyFunctionFamilyPtr createNth(const SortCtorPtr& sort_List,
                                const SortCtorPtr& sort_Int) {
    const SortVariablePtr alpha = new SortVariable;
    return new PolyFunctionFamily("nth", alpha->m(),
                                  to_vector(sort_Int->m(), sort_List->m(alpha->m())));
}

Lists::Lists(const Empty& empty, const Integers& integers)
: empty(empty), integers(integers),
  sort_List     (new SortCtor ("List", 1)),
  polyfun_nil   (new PolyFunctionFamily ("nil",
                                         sort_List->m((new SortVariable)->m()),
                                         vector<SortPtr>())),
  polyfun_cons  (createCons(sort_List)),
  polyfun_head  (createHead(sort_List)),
  polyfun_tail  (createTail(sort_List)),
  polyfun_append(createAppend(sort_List)),
  polyfun_length(new PolyFunctionFamily("length", integers.sort_Int->m(),
                                        to_vector(sort_List->m((new SortVariable)->m())))),
  polyfun_nth   (createNth(sort_List, integers.sort_Int)),
  multiArityFun_cons  (new RightAssocFunctionFamily ((FunctionFamily*)polyfun_cons)),
  multiArityFun_append(new LeftAssocFunctionFamily ((FunctionFamily*)polyfun_append))
  {}

void Lists::registerSymbolsLocally(Environment& env) const {
    env.addSortCtorFamily(sort_List->name.body,
                          new SimpleSortCtorFamily(sort_List), true);
    registerPolyFunction(env, polyfun_nil);
    registerPolyFunction(env, polyfun_head);
    registerPolyFunction(env, polyfun_tail);
    registerPolyFunction(env, polyfun_length);
    registerPolyFunction(env, polyfun_nth);
    registerMultiArityFunction(env, polyfun_cons, multiArityFun_cons);
    registerMultiArityFunction(env, polyfun_append, multiArityFun_append);
}

void Lists::registerSymbols(Environment& env) const {
    integers.registerSymbolsLocally(env);
    empty.registerSymbolsLocally(env);
    registerSymbolsLocally(env);
}

void Lists::visit(const FunctionAppPtr& node,
                  FunctionVisitor& visitor) const {
    if (!(visitLocally(node, visitor) ||
          empty.visitLocally(node, visitor) ||
          integers.visitLocally(node, visitor)))
        visitor.visit_other(node);
}

bool Lists::visitLocally(const FunctionAppPtr& node,
                         FunctionVisitor& visitor) const {
    const ParentMonoFunctionPtr parentFun = (ParentMonoFunctionPtr)node->function;
    if (!parentFun)
        return false;
    const FunctionFamilyPtr parent = parentFun->parent;
    if (parent == polyfun_nil)
        visitor.visit_polyfun_nil(node);
    else if (parent == polyfun_cons)
        visitor.visit_polyfun_cons(node);
    else if (parent == polyfun_head)
        visitor.visit_polyfun_head(node);
    else if (parent == polyfun_tail)
        visitor.visit_polyfun_tail(node);
    else if (parent == polyfun_append)
        visitor.visit_polyfun_append(node);
    else if (parent == polyfun_length)
        visitor.visit_polyfun_length(node);
    else if (parent == polyfun_nth)
        visitor.visit_polyfun_nth(node);
    else
        return false;
    return true;
}

Lists* Lists::create() {
    Empty* empty = new Empty;
    Integers* integers = new Integers (*empty);
    return new Lists (*empty, *integers);
}

const Empty& Lists::getEmpty() const {
    return empty;
}

}
