/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "maps.H"
#include "../util.H"
#include "function-visitor.H"

namespace Theories {

using namespace SMTLibAbsyn;
using namespace std;

PolyFunctionFamilyPtr createApply(const SortCtorPtr& sort_Map) {
    const SortVariablePtr alpha = new SortVariable;
    const SortVariablePtr beta = new SortVariable;
    return new PolyFunctionFamily("apply", beta->m(),
                                  to_vector(sort_Map->m(alpha->m(), beta->m()),
                                            alpha->m()));
}

PolyFunctionFamilyPtr createOverwrite(const SortCtorPtr& sort_Map) {
    const SortVariablePtr alpha = new SortVariable;
    const SortVariablePtr beta = new SortVariable;
    const SortPtr Map_alpha_beta = sort_Map->m(alpha->m(), beta->m());
    return new PolyFunctionFamily("overwrite", Map_alpha_beta,
                                  to_vector(Map_alpha_beta, alpha->m(), beta->m()));
}

PolyFunctionFamilyPtr createDomain(const SortCtorPtr& sort_Map,
                                   const SortCtorPtr& sort_Set) {
    const SortVariablePtr alpha = new SortVariable;
    const SortVariablePtr beta = new SortVariable;
    return new PolyFunctionFamily("domain", sort_Set->m(alpha->m()),
                                  to_vector(sort_Map->m(alpha->m(), beta->m())));
}

PolyFunctionFamilyPtr createRange(const SortCtorPtr& sort_Map,
                                  const SortCtorPtr& sort_Set) {
    const SortVariablePtr alpha = new SortVariable;
    const SortVariablePtr beta = new SortVariable;
    return new PolyFunctionFamily("range", sort_Set->m(beta->m()),
                                  to_vector(sort_Map->m(alpha->m(), beta->m())));
}

PolyFunctionFamilyPtr createRestriction(const string& name,
                                        const SortCtorPtr& sort_Map,
                                        const SortCtorPtr& sort_Set) {
    const SortVariablePtr alpha = new SortVariable;
    const SortVariablePtr beta = new SortVariable;
    const SortPtr Map_alpha_beta = sort_Map->m(alpha->m(), beta->m());
    return new PolyFunctionFamily(name, Map_alpha_beta,
                                  to_vector(Map_alpha_beta, sort_Set->m(alpha->m())));
}

Maps::Maps(const Empty& empty, const Sets& sets)
: empty(empty), sets(sets),
  sort_Map         (new SortCtor ("Map", 2)),
  polyfun_emptyMap (new PolyFunctionFamily ("emptyMap",
                                            sort_Map->m((new SortVariable)->m(),
                                                        (new SortVariable)->m()),
                                            vector<SortPtr>())),
  polyfun_apply    (createApply(sort_Map)),
  polyfun_overwrite(createOverwrite(sort_Map)),
  polyfun_domain   (createDomain(sort_Map, sets.sort_Set)),
  polyfun_range    (createRange(sort_Map, sets.sort_Set)),
  polyfun_restrict (createRestriction("restrict", sort_Map, sets.sort_Set)),
  polyfun_subtract (createRestriction("subtract", sort_Map, sets.sort_Set))
  {}

void Maps::registerSymbolsLocally(Environment& env) const {
    env.addSortCtorFamily(sort_Map->name.body,
                          new SimpleSortCtorFamily(sort_Map), true);
    registerPolyFunction(env, polyfun_emptyMap);
    registerPolyFunction(env, polyfun_apply);
    registerPolyFunction(env, polyfun_overwrite);
    registerPolyFunction(env, polyfun_domain);
    registerPolyFunction(env, polyfun_range);
    registerPolyFunction(env, polyfun_restrict);
    registerPolyFunction(env, polyfun_subtract);
}

void Maps::registerSymbols(Environment& env) const {
    empty.registerSymbolsLocally(env);
    sets.registerSymbolsLocally(env);
    sets.integers.registerSymbolsLocally(env);
    registerSymbolsLocally(env);
}

void Maps::visit(const FunctionAppPtr& node,
                 FunctionVisitor& visitor) const {
    if (!(visitLocally(node, visitor) ||
          empty.visitLocally(node, visitor) ||
          sets.integers.visitLocally(node, visitor) ||
          sets.visitLocally(node, visitor)))
        visitor.visit_other(node);
}

bool Maps::visitLocally(const FunctionAppPtr& node,
                        FunctionVisitor& visitor) const {
    const ParentMonoFunctionPtr parentFun = (ParentMonoFunctionPtr)node->function;
    if (!parentFun)
        return false;
    const FunctionFamilyPtr parent = parentFun->parent;
    if (parent == polyfun_emptyMap)
        visitor.visit_polyfun_emptyMap(node);
    else if (parent == polyfun_apply)
        visitor.visit_polyfun_apply(node);
    else if (parent == polyfun_overwrite)
        visitor.visit_polyfun_overwrite(node);
    else if (parent == polyfun_domain)
        visitor.visit_polyfun_domain(node);
    else if (parent == polyfun_range)
        visitor.visit_polyfun_range(node);
    else if (parent == polyfun_restrict)
        visitor.visit_polyfun_restrict(node);
    else if (parent == polyfun_subtract)
        visitor.visit_polyfun_subtract(node);
    else
        return false;
    return true;
}

Maps* Maps::create() {
    Empty* empty = new Empty;
    Integers* integers = new Integers (*empty);
    Sets* sets = new Sets (*empty, *integers);
    return new Maps (*empty, *sets);
}

const Empty& Maps::getEmpty() const {
    return empty;
}

}
