/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sstream>

#include "poly-function-family.H"
#include "../util.H"
#include "../sort-unification.H"

namespace Theories {

using namespace SMTLibAbsyn;
using namespace Resolver;
using namespace SortUnification;
using namespace std;

set<SortVariablePtr> freeVars(const vector<SortPtr>& sorts,
                                   const SortPtr& sort) {
    set<SortVariablePtr> res;
    for (vector<SortPtr>::const_iterator it = sorts.begin();
        it != sorts.end();
        ++it) {
        insertAll(res, freeVariables(*it));
    }
    insertAll(res, freeVariables(sort));

    return res;
}

PolyFunctionFamily::PolyFunctionFamily(const IndexedString& name,
                                       const SortPtr& resultSort,
                                       const std::vector<SortPtr>& argumentSorts)
    : sortVariables(freeVars(argumentSorts, resultSort)),
      resultSortSchema(resultSort), argumentSortSchemas(argumentSorts),
      name(name) {}

MonoFunctionPtr
    PolyFunctionFamily::lookup(const std::vector<std::string>& indexes,
                               const std::vector<SortPtr>& argumentSorts) {
    return lookup(indexes, 0, argumentSorts);
}

MonoFunctionPtr
    PolyFunctionFamily::lookup(const std::vector<std::string>& indexes,
                               const SortPtr& resSort,
                               const std::vector<SortPtr>& argumentSorts) {
    if (name.indexes != indexes ||
        argumentSortSchemas.size() != argumentSorts.size())
        throw NoSuchSymbolException("");

    Substitution unifier;
    for (unsigned int i = 0; i < argumentSorts.size(); ++i) {
        if (!unify(argumentSortSchemas[i], argumentSorts[i], unifier))
            throw NoSuchSymbolException("");
    }

    if (resSort && !unify(resultSortSchema, resSort, unifier))
        throw NoSuchSymbolException("");

    for (set<SortVariablePtr>::const_iterator it = sortVariables.begin();
         it != sortVariables.end();
         ++it)
        if (unifier.find(*it) == unifier.end()) {
            // this means that not all sort variables could be instantiated and
            // more information is necessary
            stringstream ss;
            ss << "Reference to function " << name << " is ambiguous.\n"
               << "Possibly you have forgotten a cast (as ...).";
            throw AmbiguityException(ss.str());
        }

    assert(sortVariables.size() == unifier.size());

    std::map<Substitution, MonoFunctionPtr>::iterator oldInstance =
        monoFunctionCache.find(unifier);
    if (oldInstance != monoFunctionCache.end())
        return oldInstance->second;

    const SortPtr resultSort =
        resSort ? resSort : substitute(resultSortSchema, unifier);

    const MonoFunctionPtr monoFunction =
        new ParentMonoFunction (name, resultSort,
                                argumentSorts.begin(), argumentSorts.end(),
                                this);
    monoFunctionCache[unifier] = monoFunction;

    return monoFunction;
}

////////////////////////////////////////////////////////////////////////////////

MonoFunctionPtr
    ParentMonoFunction::lookupNewInstance(const vector<SortPtr>& newArgumentSorts) const {
    return parent->lookup(name.indexes, newArgumentSorts);
}

MonoFunctionPtr
    ParentMonoFunction::lookupNewInstance(const SortPtr& newResSort,
                                          const vector<SortPtr>& newArgumentSorts) const {
    return parent->lookup(name.indexes, newResSort, newArgumentSorts);
}


}
