/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * The theory of finite relations with cardinality. Because this includes tuples
 * and multi-ary sort constructors and functions, we have to define our own
 * families of symbols here
 */

#ifndef THEORY_RELATIONS
#define THEORY_RELATIONS 1

#include "sets.H"
#include "poly-function-family.H"

#include <map>
#include <vector>

namespace Theories {

using namespace SMTLibAbsyn;

class TupleSortCtor : public SortCtorFamily {
private:
    std::map<unsigned int, SortCtorPtr> instances;

public:
    SortCtorPtr lookup(const std::vector<std::string>& indexes);

    bool isTupleSortCtor(const SortCtorPtr& op);
};

typedef CountingPtr<TupleSortCtor> TupleSortCtorPtr;

////////////////////////////////////////////////////////////////////////////////

class TupleFunctionFamily : public FunctionFamily {
protected:
    // We just use a vector to store the created instances
    // TODO: replace this with a more efficient map
    std::vector<MonoFunctionPtr> instances;

    const TupleSortCtorPtr sort_Tuple;

    MonoFunctionPtr lookupInstance(const std::vector<SortPtr>& argumentSorts);

public:
    TupleFunctionFamily(const TupleSortCtorPtr& sort_Tuple);

    virtual MonoFunctionPtr
        lookup(const std::vector<std::string>& indexes,
               const std::vector<SortPtr>& argumentSorts) = 0;
    MonoFunctionPtr
        lookup(const std::vector<std::string>& indexes,
               const SortPtr& resSort,
               const std::vector<SortPtr>& argumentSorts);
};

////////////////////////////////////////////////////////////////////////////////

class TupleCtor : public TupleFunctionFamily {
public:
    TupleCtor(const TupleSortCtorPtr& sort_Tuple);

    MonoFunctionPtr
        lookup(const std::vector<std::string>& indexes,
               const std::vector<SortPtr>& argumentSorts);
};

typedef CountingPtr<TupleCtor> TupleCtorPtr;

////////////////////////////////////////////////////////////////////////////////

class TupleProjection : public TupleFunctionFamily {
public:
    TupleProjection(const TupleSortCtorPtr& sort_Tuple);

    MonoFunctionPtr
        lookup(const std::vector<std::string>& indexes,
               const std::vector<SortPtr>& argumentSorts);
};

typedef CountingPtr<TupleProjection> TupleProjectionPtr;

////////////////////////////////////////////////////////////////////////////////

class TupleProduct : public TupleFunctionFamily {
private:
    const SortCtorPtr sort_Set;

public:
    TupleProduct(const TupleSortCtorPtr& sort_Tuple,
                 const SortCtorPtr& sort_Set);

    MonoFunctionPtr
        lookup(const std::vector<std::string>& indexes,
               const std::vector<SortPtr>& argumentSorts);
};

typedef CountingPtr<TupleProduct> TupleProductPtr;

////////////////////////////////////////////////////////////////////////////////

class Relations : public TheoryHelpers {
public:
    Relations(const Empty& empty, const Sets& sets);

    const Empty& empty;
    const Sets& sets;

    const TupleSortCtorPtr      sort_Tuple;
    const TupleCtorPtr          polyfun_tuple;
    const TupleProjectionPtr    polyfun_project;
    const TupleProductPtr       polyfun_product;

    void registerSymbols(Environment& env) const;
    void registerSymbolsLocally(Environment& env) const;
    void visit(const FunctionAppPtr& node, FunctionVisitor& visitor) const;
    bool visitLocally(const FunctionAppPtr& node, FunctionVisitor& visitor) const;

    const Empty& getEmpty() const;
    static Relations* create();
};

}

#endif
